/*
 * dcdc-utils - dcdc-prog.c
 *
 * Copyright (C) 2014  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ltypes.h>
#include <pu-lib.h>

#ifndef PATH_MAX
#define PATH_MAX	4096
#endif

int verbose_level;
static unsigned int rom_size = 8 * 1024;

static const char *proc_base = "/proc/ar3c/dcdc";

static const char *device = "01:0c.0";
static const char *image;
static int serial = -1;
static int gain = -1;

struct chip {
	const char *model;
	unsigned int vendor_id;
	unsigned int device_id;
	unsigned int page_size;
	unsigned int num_pages;
	unsigned int interface;
};

#define VENDOR_ATMEL	0x1e

#define CHIP(_model, _vendor_id, _device_id, _page_size, _num_pages, _iface) \
	{ \
		.model = (_model), \
		.vendor_id = (_vendor_id), \
		.device_id = (_device_id), \
		.page_size = (_page_size), \
		.num_pages = (_num_pages), \
		.interface = (_iface), \
	}

static struct chip chips[] = {
	CHIP("ATmega8", VENDOR_ATMEL, 0x9307, 64, 128, 0),
	{ NULL },
};

static const char copyright_string[] =
"dcdc-prog " PACKAGE_VERSION ".\n"
"Copyright (C) 2014  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_info(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [--] [FILE]\n"
"Options:\n"
"  -d, --device=DEVICE	use DEVICE device, \"01:0c.0\" by default\n"
"  -g, --gain=GAIN	set ADC gain multiplier, 0-65535\n"
"  -h, --help		show this help\n"
"  -q, --quiet		quiet mode\n"
"  -s, --serial=SERIAL	set serial number, 0-65535\n"
"  -v, --verbose		verbose mode\n"
"  -V, --version		show version\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_info();

	printf(usage, name ? name : "dcdc-prog");
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"device", 1, 0, 'd'},
		{"gain", 1, 0, 'g'},
		{"help", 0, 0, 'h'},
		{"quiet", 0, 0, 'q'},
		{"serial", 1, 0, 's'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "d:g:hqs:vV", long_options,
				&option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'd':
			device = strdup(optarg);
			break;
		case 'g':
			gain = strtol(optarg, NULL, 0);
			break;
		case 'h':
			show_usage(args[0]);
			break;
		case 'q':
			verbose_level--;
			break;
		case 's':
			serial = atoi(optarg);
			break;
		case 'v':
			verbose_level++;
			break;
		case 'V':
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind < argc) {
		image = args[optind++];
		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters detected."
					"\n");
	}
}

static unsigned char image_buf[65536];
static unsigned int image_size;

void image_write(unsigned int addr, u8 data)
{
	if (addr >= 65536)
		abort();
	if (addr >= rom_size) {
		fprintf(stderr, "error: address %d is out of ROM\n", addr);
		exit(EXIT_FAILURE);
	}

	image_buf[addr] = data;
	if (addr + 1 > image_size)
		image_size = addr + 1;
}

int hex_to_int(char x)
{
	if (x >= '0' && x <= '9')
		return x - '0';
	if (x >= 'A' && x <= 'F')
		return x - 'A' + 10;
	if (x >= 'a' && x <= 'f')
		return x - 'a' + 10;
	return -1;
}

int parse_hex_byte(const char *x)
{
	int lo = hex_to_int(x[1]);
	int hi = hex_to_int(x[0]);

	if (lo < 0 || hi < 0)
		return -1;

	return (hi << 4) | lo;
}

static void load_image(const char *path)
{
	FILE *f;
	char *line = NULL;
	size_t line_size = 0;
	ssize_t len;
	int data_len;
	int cmd;
	int addr_lo;
	int addr_hi;
	unsigned long addr;
	unsigned char x;
	unsigned int i;

	f = fopen(path, "r");
	if (f == NULL) {
		perror("fopen");
		exit(EXIT_FAILURE);
	}

	while (!feof(f)) {
		len = getline(&line, &line_size, f);

		if (len < 2)
			break;

		if (line[0] != ':') {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		data_len = parse_hex_byte(line + 1);
		if (data_len < 0) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		if (2 * data_len + 2 + 4 * 2 > len) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		cmd = parse_hex_byte(line + 7);
		if (cmd < 0) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		if (cmd == 1)
			break;

		if (cmd != 0) {
			fprintf(stderr, "unsupported IHEX command %d\n", cmd);
			break;
		}

		addr_hi = parse_hex_byte(line + 3);
		addr_lo = parse_hex_byte(line + 5);
		if (addr_hi < 0 || addr_lo < 0) {
			fprintf(stderr, "corrupted image: invalid addr\n");
			break;
		}

		addr = (addr_hi << 8) | addr_lo;

		for (i = 0; i < data_len; i++) {
			x = parse_hex_byte(line + 9 + 2 * i);
			image_write(addr++, x);
		}
	}
}

static void dump_page(u8 *a, u8 *b) __attribute__((unused));

static void dump_page(u8 *a, u8 *b)
{
	unsigned int i;

	for (i = 0; i < 64; i++)
		fprintf(stderr, "x[%d]	%02x %02x\n", i, a[i], b[i]);
}

static struct chip *find_chip(u8 *sig, unsigned int len)
{
	u8 vendor_id;
	u16 device_id;
	struct chip *c;

	if (len != 3)
		return NULL;
	vendor_id = sig[0];
	device_id = ((u16) sig[1] << 8) | sig[2];
	for (c = &chips[0]; c->model; c++)
		if (c->vendor_id == vendor_id && c->device_id == device_id)
			return c;
	return NULL;
}

#define PROG_ENABLE		0xac
#define PROG_ENABLE2		0x53

#define CHIP_ERASE		0xac
#define CHIP_ERASE2		0x80

#define READ_FUSE_HIGH		0x58
#define READ_FUSE_HIGH2		0x08

#define WRITE_FUSE_HIGH		0xac
#define WRITE_FUSE_HIGH2	0xa8

#define READ_FUSE_LOW		0x50
#define READ_FUSE_LOW2		0x00

#define WRITE_FUSE_LOW		0xac
#define WRITE_FUSE_LOW2		0xa0

#define FUSE_WDTON		0x40
#define FUSE_EESAVE		0x08

#define FUSE_LOW_BODLEVEL	0x80
#define FUSE_LOW_BODEN		0x40
#define FUSE_LOW_SUT1		0x20
#define FUSE_LOW_SUT0		0x10

#define WRITE_EEPROM		0xc0

#define WRITE_PAGE		0x50
#define READ_PAGE		0x30

#define CONFIG_SCALER		0x0c
#define CONFIG_SERIAL		0x0e

/* atmega8 programming interface */
#define ATMEGA_READ_SIGNATURE	0x30
#define ATMEGA_READ_PROG_LO	0x20
#define ATMEGA_READ_PROG_HI	0x28
#define ATMEGA_LOAD_PROG_LO	0x40
#define ATMEGA_LOAD_PROG_HI	0x48
#define ATMEGA_WRITE_PAGE	0x4c

static void spi_xmit(int fd, const u8 *tx, u8 *rx, unsigned int n)
{
	int ret;

	ret = xwrite(fd, tx, n);
	if (ret < 0)
		die_errno("write");

	ret = xread(fd, rx, n);
	if (ret < 0)
		die_errno("read");
}

static int prog_enable(int fd)
{
	unsigned int i;
	u8 cmd[4] = { PROG_ENABLE, PROG_ENABLE2, 0, 0 };
	u8 x[4];

	for (i = 0; i < 20; i++) {
		spi_xmit(fd, cmd, x, 4);
		if (x[1] == PROG_ENABLE && x[2] == PROG_ENABLE2)
			return 0;
	}
	return -ETIMEDOUT;
}

int read_signature_at90(int fd, u8 *signature, unsigned int len)
{
	unsigned int i;

	if (len > 4)
		len = 4;

	for (i = 0; i < len; i++) {
		u8 cmd[4] = { ATMEGA_READ_SIGNATURE, 0, i, 0 };
		u8 x[4];

		spi_xmit(fd, cmd, x, 4);
		signature[i] = x[3];
	}
	return len;
}

static void chip_erase(int fd)
{
	u8 cmd[4] = { CHIP_ERASE, CHIP_ERASE2, 0, 0 };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
}

static u8 read_fuse_high(int fd)
{
	u8 cmd[4] = { READ_FUSE_HIGH, READ_FUSE_HIGH2, 0, 0 };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
	return x[3];
}

static void write_fuse_high(int fd, u8 fuse)
{
	u8 cmd[4] = { WRITE_FUSE_HIGH, WRITE_FUSE_HIGH2, 0, fuse };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
}

static u8 read_fuse_low(int fd)
{
	u8 cmd[4] = { READ_FUSE_LOW, READ_FUSE_LOW2, 0, 0 };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
	return x[3];
}

static void write_fuse_low(int fd, u8 fuse)
{
	u8 cmd[4] = { WRITE_FUSE_LOW, WRITE_FUSE_LOW2, 0, fuse };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
}

static int read_page(int fd, unsigned int page, void *buf,
		unsigned int page_size)
{
	u8 cmd[4];
	u8 x[4];
	uint8_t *data = buf;
	unsigned int i;
	unsigned int page_offset = page * page_size >> 1;

	memset(cmd, 0, sizeof(cmd));
	for (i = 0; i < page_size - 1; i += 2) {
		unsigned int addr = page_offset + (i >> 1);

		cmd[0] = ATMEGA_READ_PROG_LO;
		cmd[1] = addr >> 8;
		cmd[2] = addr & 0xff;
		spi_xmit(fd, cmd, x, 4);
		data[i + 0] = x[3];

		cmd[0] = ATMEGA_READ_PROG_HI;
		spi_xmit(fd, cmd, x, 4);
		data[i + 1] = x[3];
	}
	return 0;
}

static int write_page(int fd, unsigned int page, const void *buf,
		unsigned int page_size)
{
	u8 cmd[4];
	u8 x[4];
	const uint8_t *data = buf;
	unsigned int i;
	unsigned int page_offset = page * page_size >> 1;

	for (i = 0; i < page_size - 1; i += 2) {
		unsigned int addr = page_offset + (i >> 1);

		cmd[0] = ATMEGA_LOAD_PROG_LO;
		cmd[1] = addr >> 8;
		cmd[2] = addr & 0xff;
		cmd[3] = data[i + 0];
		spi_xmit(fd, cmd, x, 4);

		cmd[0] = ATMEGA_LOAD_PROG_HI;
		cmd[1] = addr >> 8;
		cmd[2] = addr & 0xff;
		cmd[3] = data[i + 1];
		spi_xmit(fd, cmd, x, 4);
	}
	cmd[0] = ATMEGA_WRITE_PAGE;
	cmd[1] = page_offset >> 8;
	cmd[2] = page_offset & 0xff;
	cmd[3] = 0;
	spi_xmit(fd, cmd, x, 4);
	return 0;
}

static void program_fuse(int spi_fd)
{
	u8 old;
	u8 new;
	u8 v;

	old = read_fuse_high(spi_fd);
	new = old;
	new &= ~(FUSE_WDTON | FUSE_EESAVE);
	if (old != new) {
		write_fuse_high(spi_fd, new);
		v = read_fuse_high(spi_fd);
		info("Fuse High bits: %02x -> %02x (%02x)", old, v, new);
	}

	old = read_fuse_low(spi_fd);
	new = old;
	new &= ~(FUSE_LOW_BODLEVEL | FUSE_LOW_BODEN);
	new &= ~(FUSE_LOW_SUT0 | FUSE_LOW_SUT1);
	if (old != new) {
		write_fuse_low(spi_fd, new);
		v = read_fuse_low(spi_fd);
		info("Fuse Low bits: %02x -> %02x (%02x)", old, v, new);
	}
}

#define CONFIG_BASE		0x0000

#define CONFIG_COPIES	2

static void ee_write(int fd, u8 data, unsigned int addr)
{
	u8 cmd[4] = { WRITE_EEPROM, addr >> 8, addr & 0xff, data };
	u8 x[4];

	spi_xmit(fd, cmd, x, 4);
}

static void config_write(int fd, u16 x, unsigned char reg)
{
	unsigned int addr = CONFIG_BASE + 8 * reg;
	unsigned char i;
	u16 csum;

	csum = ~x;
	for (i = 0; i < CONFIG_COPIES; i++) {
		ee_write(fd, x >> 8, addr++);
		ee_write(fd, x & 0xff, addr++);
		ee_write(fd, csum >> 8, addr++);
		ee_write(fd, csum & 0xff, addr++);
	}
}

static void do_prog(void)
{
	struct chip *c;
	unsigned int i;
	unsigned int page;
	unsigned int retry = 10;
	unsigned int to_write;
	u8 *page_buf;
	u8 *page_vbuf;
	int ret;
	u8 sig[3];
	char name[PATH_MAX];
	int state_fd;
	int spi_fd;

	snprintf(name, sizeof(name), "%s/%s/state",
			proc_base, device);
	state_fd = open(name, O_RDWR);
	if (state_fd == -1)
		die("Can't open %s: %s\n", name, strerror(errno));

	snprintf(name, sizeof(name), "%s/%s/spi",
			proc_base, device);
	spi_fd = open(name, O_RDWR);
	if (spi_fd == -1)
		die("Can't open %s: %s\n", name, strerror(errno));

	ret = xread(state_fd, name, sizeof(name));
	if ((ret == 6) && !strncmp(name, "reset", 5)) {
		info("Device already is RESET state");
		info("Deasserting RESET for a while...");
		write(state_fd, "boot\n", 5);
		usleep(10);
	}

	write(state_fd, "reset\n", 6);

	fprintf(stderr, "Enabling programming mode...");
	sleep(1);
	ret = prog_enable(spi_fd);
	if (ret) {
		fprintf(stderr, "FAILED\n");
		fprintf(stderr, "Cannot enable programming mode.\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK\n");

	fprintf(stderr, "Reading AT89S52 chip signature...");
	ret = read_signature_at90(spi_fd, sig, 3);
	if (ret != 3) {
		fprintf(stderr, "FAILED\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK (0x%02x 0x%02x 0x%02x)\n",
			sig[0], sig[1], sig[2]);

	c = find_chip(sig, 3);
	if (c == NULL) {
		fprintf(stderr, "error: unknown chip detected\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "Found %s, ROM %ld kB (page size %d)\n",
			c->model, (unsigned long) c->page_size
			* c->num_pages >> 10, c->page_size);

	page_buf = malloc(c->page_size);
	page_vbuf = malloc(c->page_size);
	if (page_buf == NULL || page_vbuf == NULL) {
		fprintf(stderr, "Cannot allocate page buffer\n");
		exit(EXIT_FAILURE);
	}

	program_fuse(spi_fd);

	if (image) {
		fprintf(stderr, "Erasing chip...");
		chip_erase(spi_fd);
		usleep(100000);
		fprintf(stderr, "OK\n");
	}

	fprintf(stderr, "Enabling programming mode...");
	ret = prog_enable(spi_fd);
	if (ret) {
		fprintf(stderr, "FAILED\n");
		fprintf(stderr, "Cannot enable programming mode.\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK\n");

	if (image) {
		fprintf(stderr, "Checking first page...");
		read_page(spi_fd, 0, page_buf, c->page_size);
		for (i = 0; i < c->page_size; i++)
			if (page_buf[i] != 0xff) {
				fprintf(stderr, "error\n");
				die("Page not erased\n");
			}
		fprintf(stderr, "OK\n");

		to_write = (image_size + c->page_size) / c->page_size;
		if (to_write >= c->num_pages)
			die("image too big: %d pages, only %d available\n",
					to_write, c->num_pages);

		for (page = 0; page < to_write; ) {
			fprintf(stderr, "programming page %d/%d\n",
					page, to_write);
			memcpy(page_buf, image_buf + page * c->page_size,
					c->page_size);
			write_page(spi_fd, page, page_buf, c->page_size);
			usleep(5000);
			read_page(spi_fd, page, page_vbuf, c->page_size);
			if (memcmp(page_vbuf, page_buf, c->page_size)) {
				fprintf(stderr, "page %d: write error\n",
						page);
				retry--;
				if (!retry)
					die("retry limit reached\n");
				continue;
			}
			page++;
		}
	}

	if (serial != -1)
		config_write(spi_fd, serial, CONFIG_SERIAL);
	if (gain != -1)
		config_write(spi_fd, gain, CONFIG_SCALER);

	write(state_fd, "boot\n", 5);
}

int main(int argc, char **args)
{
	parse_command_line(argc, args);

	if (image) {
		load_image(image);

		fprintf(stderr, "Loaded %d bytes image\n", image_size);
	}

	if (image || (serial != -1) || (gain != -1))
		do_prog();
	return 0;
}
